package atomicwheel;

import java.io.Serializable;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

/**
 * A simple node/edge attribute. It doesn't have a name, only a type -- the name
 * is only stored by the {@link AttributeManager}, if it is part of one.
 * 
 * @param <GI>
 *            the {@link GraphItem} that serves as the key
 * @param <T>
 *            type of the attribute values
 * 
 * @author matthias
 */
public class Attribute<GI extends GraphItem, T> implements Serializable {
	private static final long serialVersionUID = 1L;

	final Class<?> type;
	final boolean isList;
	final Map<GI, T> values = new HashMap<GI, T>();

	/**
	 * Creates a new attribute that is a copy of the given attribute.
	 * <p>
	 * Does not create deep copies for list attributes!
	 * 
	 * @param attr
	 *            the attribute to copy
	 */
	public Attribute(final Attribute<GI, T> attr) {
		type = attr.type;
		isList = attr.isList;
		values.putAll(attr.values);
	}

	/**
	 * Creates a new non-list attribute with the given value type.
	 * 
	 * @param type
	 *            the {@link Class} of the value type
	 */
	public Attribute(final Class<T> type) {
		this.type = type;
		isList = false;
	}

	/**
	 * Creates a new list attribute with the given value type.
	 * 
	 * @param listType
	 *            the {@link Class} of the list itsself
	 * @param elementType
	 *            the {@link Class} of the values in the list
	 */
	public Attribute(
			@SuppressWarnings("rawtypes") final Class<? extends List> listType,
			final Class<?> elementType) {
		this.type = elementType;
		isList = true;
	}

	/**
	 * @param item
	 *            item whose value to set
	 * @param value
	 *            value to set it to
	 */
	public void set(final GI item, final T value) {
		values.put(item, value);
	}

	/**
	 * Query an item's value.
	 * 
	 * @param item
	 *            item whose value to query
	 * @param defaultValue
	 *            value to return if the item doesn't have a value assigned
	 * @return the value associated with the item, or {@code defaultValue} if
	 *         none
	 */
	public T get(final GI item, final T defaultValue) {
		if (values.containsKey(item))
			return values.get(item);
		return defaultValue;
	}

	/**
	 * Query an item's value. Throws a {@link NoSuchElementException} if there
	 * is no value for that item.
	 * 
	 * @param item
	 *            item whose value to query
	 * @return the value associated with the item
	 */
	public T get(final GI item) {
		if (!values.containsKey(item))
			throw new NoSuchElementException(item.toString());
		return values.get(item);
	}

	/**
	 * @return the type of the values, or of the list elements for list
	 *         attributes
	 */
	public Class<?> getType() {
		return type;
	}

	/** @return <code>true</code> if this is a list attribute */
	public boolean isList() {
		return isList;
	}
}
