package atomicwheel;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Map.Entry;
import java.util.Set;

import atomicwheel.util.SetMap;

/**
 * A simple graph.
 * 
 * @author matthias
 */
public class Graph extends GraphItem implements Serializable {
	private static final long serialVersionUID = 1L;
	private final Set<Node> nodes = new HashSet<Node>();
	private final Set<Edge> edges = new HashSet<Edge>();
	private final AttributeManager<Node> nodeAttr = new AttributeManager<Node>();
	private final AttributeManager<Edge> edgeAttr = new AttributeManager<Edge>();
	private final AttributeManager<Graph> graphAttr = new AttributeManager<Graph>();
	private final SetMap<Node, Edge> edgeSet = new SetMap<Node, Edge>();

	/** Creates an empty graph. */
	public Graph() {
	}

	/**
	 * Creates a graph that is a copy of the given graph. The new graph has its
	 * own {@link AttributeManager}s containing independent {@link Attribute}s.
	 * 
	 * @param graph
	 *            the graph to copy
	 */
	@SuppressWarnings({ "rawtypes", "unchecked" })
	public Graph(final Graph graph) {
		this(graph.nodes, graph.edges);
		for (final Entry<String, Attribute<Node, ?>> attr : graph.nodeAttr.attributes
				.entrySet())
			nodeAttr.add(attr.getKey(), new Attribute(attr.getValue()));
		for (final Entry<String, Attribute<Edge, ?>> attr : graph.edgeAttr.attributes
				.entrySet())
			edgeAttr.add(attr.getKey(), new Attribute(attr.getValue()));
	}

	/**
	 * Creates a new graph containing the given nodes and edges.
	 * 
	 * @param nodes
	 *            nodes to include
	 * @param edges
	 *            edges to include
	 */
	public Graph(final Collection<Node> nodes, final Collection<Edge> edges) {
		for (final Node node : nodes)
			add(node);
		for (final Edge edge : edges)
			add(edge);
	}

	/**
	 * Adds the given node to this graph. This doesn't change the node in any
	 * way; it can be part of any number of graphs, including none.
	 * 
	 * @param node
	 *            node to add
	 */
	public void add(final Node node) {
		nodes.add(node);
	}

	/**
	 * Removes the given node, and all edges incident to it, from this graph.
	 * This doesn't change the edge in any way; it can be part of any number of
	 * graphs, including none.
	 * 
	 * @param node
	 *            node to remove
	 */
	public void remove(final Node node) {
		nodes.remove(node);
		for (final Edge edge : new ArrayList<Edge>(getIncidentEdges(node)))
			remove(edge);
	}

	/** @return the set of nodes */
	public Set<Node> getNodes() {
		return nodes;
	}

	/**
	 * Adds the given edge, and its endnodes, to this graph.
	 * 
	 * @param edge
	 *            edge to add
	 */
	public void add(final Edge edge) {
		edges.add(edge);
		edgeSet.add(edge.source, edge);
		edgeSet.add(edge.target, edge);
		add(edge.source);
		add(edge.target);
	}

	/**
	 * Removes the given edge from this graph.
	 * 
	 * @param edge
	 *            edge to remove
	 */
	public void remove(final Edge edge) {
		edges.remove(edge);
		edgeSet.remove(edge.source, edge);
		edgeSet.remove(edge.target, edge);
	}

	/** @return the set of edges */
	public Set<Edge> getEdges() {
		return edges;
	}

	/**
	 * Gets all edges incident with the given node, regardless of direction.
	 * 
	 * @param node
	 *            the node to query
	 * @return a set of all edges incident to it
	 */
	public Set<Edge> getIncidentEdges(final Node node) {
		return edgeSet.get(node);
	}

	/** @return the {@link AttributeManager} for nodes. */
	public AttributeManager<Node> getNodeAttributeManager() {
		return nodeAttr;
	}

	/** @return the {@link AttributeManager} for edges. */
	public AttributeManager<Edge> getEdgeAttributeManager() {
		return edgeAttr;
	}

	/** @return the {@link AttributeManager} for the entire graph. */
	public AttributeManager<Graph> getGraphAttributeManager() {
		return graphAttr;
	}
}
